"""
Cross referencing tools and methods for ``pyXMIP`` catalogs.

More details on cross referencing can be found in the user guide: :ref:`cross_Referencing`

Notes
-----

The :py:mod:`cross_reference` module is the core of the user-tools in ``pyXMIP``. This module allows you to cross reference tables,
run queries and reduce results.

"""
import pathlib as pt
import time
from typing import Annotated, Any, Callable, Collection, Self, TypeVar

import numpy as np
import pandas as pd
import sqlalchemy as sql
from astropy.coordinates import ICRS, SkyCoord
from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from pyXMIP.schema import CMDSchema
from pyXMIP.structures.databases import (
    DEFAULT_DATABASE_REGISTRY,
    DBRegistry,
    SourceDatabase,
)
from pyXMIP.structures.table import SourceTable
from pyXMIP.utilities.logging import mainlog
from pyXMIP.utilities.types import _CMDTypePydanticAnnotation

# ========================================================= #
# Type hinting constructs                                   #
# ========================================================= #
RProc = TypeVar("RProc")

# ========================================================= #
# Cross Matching Table Class                                #
# ========================================================= #


class CrossMatchDatabase:
    """
    Database of cross-matching results.

    The :py:class:`CrossMatchDatabase` class provides an easy-to-use API for accessing the results of a cross-matching
    procedure (as stored in a SQL database). Furthermore, :py:class:`CrossMatchDatabase` instances provide a variety of
    methods for sorting, processing, and performing other important tasks on the underlying data.

    Notes
    -----

    The :py:class:`CrossMatchDatabase` instance provides access to an underlying SQL database with the following 3 types
    of tables:

    1. **Catalog**: (``CATALOG``) This table contains the original catalog that was used for the cross-matching procedure.
    2. **Match Tables**: These are the results of cross matching the catalog against each of the relevant databases.
    3. **META**: The Meta table contains information on what procedures have already been performed on the database.


    """

    def __init__(
        self,
        path: str | pt.Path,
        schema: CMDSchema = None,
        overwrite_schema: bool = False,
    ):
        """
        Initialize a :py:class:`CrossMatchDatabase` instance.

        Parameters
        ----------
        path: str or :py:class:`pathlib.Path`
            The path to the underlying ``SQL`` database to load.
        schema: :py:class:`schema.CMDSchema`, optional
            The :py:class:`schema.CMDSchema` corresponding to this cross-matching database. If it is not specified, then
            the class will search for a schema file with the correct signature in the same directory as ``path``.

            .. hint::

                For a filename (as specified by ``path``), we expect to locate a schema with path ``path_schema.json``.

        overwrite_schema: bool, optional
            If ``True`` and a schema is specified (``schema != None``), then that schema will be written to disk in the
            standard position even if one already exists.

        """

        # ------------------------------------------------ #
        # Defining attributes                              #
        # ------------------------------------------------ #
        self.path: pt.Path = pt.Path(path)
        """:py:class:`pathlib.Path`: Path to the underlying SQL database.
        """
        assert self.path.exists(), f"The path {self.path.absolute()} doesn't exist."

        self._sql_engine = sql.create_engine(f"sqlite:///{self.path}")

        # -- Hidden attributes for property creation -- #
        self._tables = None

        # -- Seek schema -- #
        self._schema, self.has_schema, self.schema_path = self._get_schema(
            schema=schema, overwrite_schema=overwrite_schema
        )

    def _get_schema(self, schema: CMDSchema = None, overwrite_schema: bool = False):
        """

        Parameters
        ----------
        schema
        overwrite_schema

        Returns
        -------

        """
        _schema_path = pt.Path(str(self.path.with_suffix("")) + "_schema.json")
        if schema is not None:
            # The schema was specified, we don't need to do anything.
            schema: CMDSchema = schema
            has_schema: bool = True

            if (_schema_path.exists() and overwrite_schema) or (
                not _schema_path.exists()
            ):
                # We need to replace the schema path.
                schema.write(
                    _schema_path, overwrite=overwrite_schema, file_format="json"
                )

        else:
            # We need to search for one.
            if _schema_path.exists():
                # the schema file exists and we can load it.
                schema = CMDSchema.read(_schema_path, file_format="json")
                has_schema = True
            else:
                mainlog.warning(
                    f"Failed to load CMDSchema with CrossMatchDatabase instance {self}."
                )
                schema = None
                has_schema = False

        return schema, has_schema, _schema_path

    def __getitem__(self, item: str) -> pd.DataFrame:
        with self.connect() as conn:
            return pd.read_sql_table(item, conn)

    def __str__(self):
        return f"<CrossMatchDatabase @ {self.path}>"

    def __repr__(self):
        return self.__str__()

    def __contains__(self, item: str | SourceDatabase) -> bool:
        if isinstance(item, str):
            return f"{item}_MATCH" in self.match_tables
        else:
            return f"{item.name}_MATCH" in self.match_tables

    @property
    def schema(self) -> CMDSchema | None:
        if self.has_schema:
            return self._schema
        else:
            mainlog.warning(
                f"There is no schema associated with {self}. You may need to set / create one."
            )
            return None

    @schema.setter
    def schema(self, value: CMDSchema):
        self._schema, self.has_schema, _ = self._get_schema(schema=value)

    def overwrite_schema(self):
        self._schema, self.has_schema, self.schema_path = self._get_schema(
            schema=self.schema, overwrite_schema=True
        )

    @property
    def tables(self) -> list[str]:
        """list of str: List of available tables.

        See Also
        --------
        drop_table, match_tables
        """
        if self._tables is None:
            _insp = sql.inspect(self._sql_engine)
            table_names = _insp.get_table_names()
            self._tables = table_names
        return self._tables

    @property
    def meta(self) -> pd.DataFrame:
        """:py:class:`pd.DataFrame`: Dataframe representing the meta-processes that have operated on this instance.

        The ``META`` table (within the larger SQL database) is a repository of the "edit history" of the database. Effectively,
        different processes can check ``META`` to determine if certain tasks have already been performed or if critical
        pre-requisite tasks have actually occurred.

        See Also
        --------
        meta_reset, meta_remove, meta_add, check_meta, build_meta_table

        """
        if "META" in self.tables:
            pass
        else:
            self.build_meta_table(overwrite=True)

        with self._sql_engine.connect() as conn:
            return pd.read_sql_table("META", conn)

    @property
    def has_catalog(self) -> bool:
        """bool: Returns ``True`` if the catalog is already loaded.

        If it's loaded, the original source catalog ends up in the ``CATALOG`` table of the SQL database.

        See Also
        --------
        add_catalog, add_catalog_from_table

        """
        return "CATALOG" in self.tables

    @property
    def match_tables(self) -> list[str]:
        """list of str: List of the available match tables.

        .. hint::

            Match tables are tables in the underlying SQL database which are actually cross-matching database.

        See Also
        --------
        tables, meta

        """
        return [i for i in self.tables if i[-6:] == "_MATCH"]

    def _reset_attributes(self):
        self._tables = None

    def drop_table(self, table_name: str):
        """
        Drop the table from the ``SQL`` database.

        Parameters
        ----------
        table_name: str
            The name of the table to delete.

        Returns
        -------
        None

        See Also
        --------
        tables, match_tables, cross_match
        """
        assert table_name in self.tables, f"Table {table_name} doesn't exist."

        with self._sql_engine.connect() as conn:
            query = sql.text(f"DROP TABLE {table_name}")
            conn.execute(query)

        mainlog.info(f"DELETED table {table_name} from {self.path}.")

    def query(self, query: str) -> pd.DataFrame:
        """
        Query the SQL database.

        Parameters
        ----------
        query: str
            The SQLITE flavor SQL query for this database.

        """
        with self._sql_engine.connect() as conn:
            return pd.read_sql_query(sql.text(query), conn)

    def cross_match(
        self,
        databases: list[SourceDatabase | str],
        registry: DBRegistry = None,
        **kwargs,
    ):
        """
        Add a new cross-matching table to this object.

        Parameters
        ----------
        databases: list of str or :py:class:`structures.databases.SourceDatabase`
            The database to cross-reference against.
        registry: :py:class:`structures.databases.DBRegistry`, optional
            The database registry to use. If unspecified, then the default registry is used.

        """
        from pyXMIP.structures.table import SourceTable

        # ================================================= #
        # Setup: args / kwargs                              #
        # ================================================= #
        if not self.has_schema:
            mainlog.warning(
                f"{self} doesn't have a schema. The catalog will generate it's own (possible erroneous) schema."
            )

        if registry is None:
            registry = DEFAULT_DATABASE_REGISTRY

        for k, database in enumerate(databases):
            if isinstance(database, str):
                # this is a string database and needs to be looked up.
                try:
                    databases[k] = registry[database]
                except KeyError:
                    raise ValueError(
                        f"The database string {database} was not found in the registry. Have you added the database to your registry?"
                    )
            else:
                # This database is a class of its own.
                pass

        if any(database in self for database in databases):
            mainlog.warning(
                f"Databases {[database.name for database in databases if database in self]} are already in this cross-match database. Delete them and re-run to overwrite."
            )
            databases = [database for database in databases if database not in self]

        mainlog.info(f"Adding cross-match results from {databases} to {self.path}.")

        # -- load the catalog as a source table -- #
        assert (
            self.has_catalog
        ), f"There is no CATALOG object in {self}. Add a catalog to proceed."

        with self.connect() as conn:
            _catalog = SourceTable.from_pandas(pd.read_sql_table("CATALOG", conn))

        if self.has_schema:
            _catalog.schema = self.schema.catalog_schema
        else:
            _ = (
                _catalog.schema
            )  # This just base loads the schema to make sure it generates.

        # ================================================= #
        # Cross Referencing                                 #
        # ================================================= #
        cross_match_table(_catalog, self.path, databases=databases, **kwargs)

    def build_meta_table(self, overwrite: bool = False):
        """
        Generate a blank version of the ``META`` table in the database.

        Parameters
        ----------
        overwrite: bool
            If ``True``, overwrites will be allowed.

        See Also
        --------
        meta, check_meta, meta_add, meta_remove, meta_reset
        """
        # ========================================= #
        # Setting up the procedure                  #
        # ========================================= #
        if "META" in self.tables:
            if overwrite:
                with self._sql_engine.connect() as conn:
                    conn.execute(sql.text("DROP TABLE META"))
            else:
                raise ValueError(
                    "Failed to generate new META table because META already exists and overwrite=False."
                )
        else:
            pass

        # ======================================== #
        # Building the table                       #
        # ======================================== #
        tbl = {
            "PROCESS": ["META_GENERATED"],
            "TABLE": ["ALL"],
            "DATE_RUN": [time.asctime()],
            "REDUCTION": [False],
        }

        pd.DataFrame(tbl).to_sql(
            "META", self._sql_engine, if_exists="replace", index=False
        )

    def meta_add(self, process: str, table: str, is_reduction: bool = False):
        """
        Add a record to the ``META`` table.

        Parameters
        ----------
        process: str
            The process to add.
        table: str
            The table to add.
        is_reduction: bool
            Mark this additional process as a reduction process.

            .. note::

                If a process is marked as a reduction process, then it should produce a column in the applied table
                called ``<process_name>_SCORE``.

        See Also
        --------
        meta, check_meta, build_meta_table, meta_remove, meta_reset
        """
        mainlog.debug(f"Added {process} flag to {self.path} for {table}.")
        tbl = {
            "PROCESS": [process],
            "TABLE": [table],
            "DATE_RUN": [time.asctime()],
            "REDUCTION": [is_reduction],
        }

        pd.DataFrame(tbl).to_sql(
            "META", self._sql_engine, index=False, if_exists="append"
        )
        self._reset_attributes()

    def meta_remove(self, process: str, table: str):
        """
        Remove a record from the ``META`` table.

        Parameters
        ----------
        process: str
            The process to remove.
        table: str
            The table to remove.

        See Also
        --------
        meta, check_meta, meta_add, build_meta_table, meta_reset

        """
        _new_meta = self.meta.loc[
            (self.meta["PROCESS"] != process) & (self.meta["TABLE"] != table), :
        ]

        _new_meta.to_sql("META", self._sql_engine, if_exists="replace", index=False)
        self._reset_attributes()

    def meta_reset(self):
        """
        Reset the ``META`` table.

        See Also
        --------
        meta, check_meta, meta_add, meta_remove, build_meta_table
        """
        self.build_meta_table(overwrite=True)
        self._reset_attributes()

    def check_meta(self, process: str, table: str) -> bool:
        """
        Check the ``META`` table for a record.

        Parameters
        ----------
        process: str
            The process to check for.
        table: str
            The table to check for.

        Returns
        -------
        bool

        See Also
        --------
        meta, check_meta, meta_add, meta_remove, meta_reset
        """
        return (
            len(
                self.meta.loc[
                    (self.meta["PROCESS"] == process) & (self.meta["TABLE"] == table), :
                ]
            )
            != 0
        )

    def get_database(
        self, table_name: str, registry: DBRegistry = None
    ) -> SourceDatabase:
        """
        Given a ``table_name`` in this cross-matching database, determines the database from which it originated.

        Parameters
        ----------
        table_name: str
            The table in the SQL file for which to seek a matching database.
        registry: DBRegistry, optional
            The database registry to search for the databases. If ``None`` (default), then the default registry is used.

            .. important::

                If you have custom databases that were used to build the cross-matching database, you'll need to make sure
                they're in the relevant registry or this process will fail.

        Returns
        -------
        SourceDatabase
            The database matching the desired table.

        Raises
        ------
        ValueError
            Raised if the process fails to locate a database with a name matching the table name.
        """
        if registry is None:
            registry = DEFAULT_DATABASE_REGISTRY
        # -- First, check that the table is actually valid -- #
        assert (
            table_name in self.match_tables
        ), f"The table {table_name} is not among the tables of this database or is not a match table."

        # -- Pull the database name -- #
        db_name = table_name[:-6]  # Removes _MATCH from the end of the name string.

        # -- Search for a database -- #
        for name in registry.names:
            if name == db_name:
                return registry[name]

        raise ValueError(
            f"Failed to find a database matching {db_name} in the provided registry. Have you made sure all of the relevant databases are in the registry?"
        )

    @staticmethod
    def chunk_db_operation(
        flag: str,
        afunc: Callable[[Any, str, str], tuple[Collection[Any], dict]],
        allow_overwrite: bool = False,
        **meta_kwargs,
    ) -> Callable[[Callable], Callable]:
        """
        Meta-decorator for wrapping base processes and simplifying runtime context to simple operation on table.

        Parameters
        ----------
        flag: str
            The process flag to assign to the decorated process / method. This should be unique. It is what META uses to
            identify if a process has already been run or not.
        afunc: Callable
            A function (staticmethod) which takes an ``instance``, ``table`` and ``process_flag`` (along with ``**kwargs``) and returns
            ``args`` and ``kwargs`` which can then be passed directly to the decorated method.
        allow_overwrite: bool, optional
            If ``True``, then overwriting is permitted on this operation.
        """
        from pyXMIP.utilities.sql import chunk_sql_query_operation

        def _chunk_db_operation(function: Callable[[pd.DataFrame, ...], pd.DataFrame]):
            """
            Decorator function to perform operations on SQL queries in chunks instead of loading the entire database into
            memory at once.

            Parameters
            ----------
            function: Callable
                The operation to perform on each of the query chunks.

                The signature of ``function`` should be ``func(table: pd.DataFrame, *args, **kwargs)``. ``*args`` and ``**kwargs`` are
                then provided at runtime by the user. (See the notes on the returned function)

            Returns
            -------
            Callable
                The resulting decorator output.

                This is a function with signature ``func(engine: sql.Engine, sql_query: str, otable_name: str,chunksize:int = 1000, *args, **kwargs)``.
                The ``engine`` must be the SQL engine to connect to, the ``sql_query`` is the query to execute and the result is then operated on
                by the wrapped function. Finally, ``otable_name`` specifies what name to provide to the output table. ``*args,**kwargs`` are passed
                directly to ``function``.

            """

            def wrapper(
                self: Self,
                sql_query: str,
                otable_name: str,
                *args,
                **kwargs,
            ) -> None:
                """
                Perform a chunkwise operation on a particular ``sql_query`` and write the resulting output to disk.

                Parameters
                ----------
                self: CrossMatchDatabase
                    The instance to perform this operation.
                sql_query: str
                    The sql query to run.
                otable_name: str
                    The name to give to the output table.
                chunksize: int, optional
                    The maximum allowed chunksize for each operation.
                overwrite: bool, optional
                    If ``True``, then the process will proceed regardless of the META check.
                args:
                    Additional arguments to pass to the underlying function.
                kwargs:
                    Additional keyword arguments to pass to the underlying function.
                """
                # -- Managing the overwrite protocol -- #
                # -- Setup -- #
                # fix the function signature.
                f = lambda *a, **k: function(self, *a, **k)

                # add TQDM params to meta_kwargs if not specified.
                if "tqdm_kwargs" not in meta_kwargs:
                    meta_kwargs["tqdm_kwargs"] = {}

                meta_kwargs["tqdm_kwargs"]["desc"] = meta_kwargs["tqdm_kwargs"].get(
                    "desc", f"[{flag}]"
                )
                _base_wrapper = chunk_sql_query_operation(**meta_kwargs)(f)

                overwrite = kwargs.pop("overwrite", False)
                if overwrite and not allow_overwrite:
                    mainlog.warning(
                        f"[{flag}@{otable_name}] Operation doesn't permit overwriting. Overwrite=False."
                    )
                    overwrite = False

                # -- perform the argument generator process -- #
                # We run the argument fetching function for this process (flag) over unspecified tables (this is a general process).
                if afunc is not None:
                    _args, _kwargs = afunc(self, flag, otable_name)

                    args = list(args) + list(_args)
                    kwargs = {
                        **kwargs,
                        **{k: v for k, v in _kwargs.items() if k not in kwargs},
                    }
                else:
                    pass

                # -- perform operation checks -- #
                operation_check = self.check_meta(flag, otable_name)

                if operation_check and not overwrite:
                    mainlog.warning(
                        f"[{flag}@{otable_name}] Failed to execute because it is already in META and overwrite=False."
                    )
                    return None
                elif operation_check and overwrite:
                    mainlog.info(
                        f"[{flag}@{otable_name}] This process has already been run, but overwrite=True so it has been permitted."
                    )
                else:
                    pass

                engine = self._sql_engine

                _base_wrapper(*(engine, sql_query, otable_name, *args), **kwargs)

                self.meta_add(flag, otable_name)

            return wrapper

        return _chunk_db_operation

    @staticmethod
    def chunk_db_table_operation(
        flag: str,
        afunc: Callable[[Any, str, str], tuple[Collection[Any], dict]],
        allow_overwrite: bool = False,
        **meta_kwargs,
    ) -> Callable[[Callable], Callable]:
        """
        Meta-decorator for wrapping base processes and simplifying runtime context to simple operation on table.

        Parameters
        ----------
        flag: str
            The process flag to assign to the decorated process / method. This should be unique. It is what META uses to
            identify if a process has already been run or not.
        afunc: Callable
            A function (staticmethod) which takes an ``instance``, ``table`` and ``process_flag`` (along with ``**kwargs``) and returns
            ``args`` and ``kwargs`` which can then be passed directly to the decorated method.
        allow_overwrite: bool, optional
            If ``True``, then overwriting is permitted on this operation.
        """
        from pyXMIP.utilities.sql import chunk_sql_table_operation

        def _chunk_db_table_operation(
            function: Callable[[pd.DataFrame, ...], pd.DataFrame]
        ):
            """
            Decorator function to perform operations on SQL queries in chunks instead of loading the entire database into
            memory at once.

            Parameters
            ----------
            function: Callable
                The operation to perform on each of the query chunks.

                The signature of ``function`` should be ``func(table: pd.DataFrame, *args, **kwargs)``. ``*args`` and ``**kwargs`` are
                then provided at runtime by the user. (See the notes on the returned function)

            Returns
            -------
            Callable
                The resulting decorator output.

                This is a function with signature ``func(engine: sql.Engine, sql_query: str, otable_name: str,chunksize:int = 1000, *args, **kwargs)``.
                The ``engine`` must be the SQL engine to connect to, the ``sql_query`` is the query to execute and the result is then operated on
                by the wrapped function. Finally, ``otable_name`` specifies what name to provide to the output table. ``*args,**kwargs`` are passed
                directly to ``function``.

            """

            def wrapper(
                self: Self,
                table: str,
                *args,
                **kwargs,
            ) -> None:
                """
                Perform a chunkwise operation on a particular ``sql_query`` and write the resulting output to disk.

                Parameters
                ----------
                self: CrossMatchDatabase
                    The instance to perform this operation.
                chunksize: int, optional
                    The maximum allowed chunksize for each operation.
                overwrite: bool, optional
                    If ``True``, then the process will proceed regardless of the META check.
                args:
                    Additional arguments to pass to the underlying function.
                kwargs:
                    Additional keyword arguments to pass to the underlying function.
                """
                # -- Setup -- #
                # fix the function signature.
                f = lambda *a, **k: function(self, *a, **k)

                # add TQDM params to meta_kwargs if not specified.
                if "tqdm_kwargs" not in meta_kwargs:
                    meta_kwargs["tqdm_kwargs"] = {}
                meta_kwargs["tqdm_kwargs"]["desc"] = f"[{flag}@{table}]"

                _base_wrapper = chunk_sql_table_operation(**meta_kwargs)(f)

                # -- Managing the overwrite protocols -- #
                overwrite = kwargs.pop("overwrite", False)
                if overwrite and not allow_overwrite:
                    mainlog.warning(
                        f"[{flag}@{table}] Operation doesn't permit overwriting. Overwrite=False."
                    )
                    overwrite = False

                # -- perform the argument generator process -- #
                # We run the argument fetching function for this process (flag) over unspecified tables (this is a general process).
                if afunc is not None:
                    _args, _kwargs = afunc(self, flag, table)

                    args = list(_args) + list(args)
                    kwargs = {
                        **kwargs,
                        **{k: v for k, v in _kwargs.items() if k not in kwargs},
                    }

                else:
                    pass

                # -- perform operation checks -- #
                operation_check = self.check_meta(flag, table)

                if operation_check and not overwrite:
                    mainlog.warning(
                        f"[{flag}@{table}] Failed to execute because it is already in META and overwrite=False."
                    )
                    return None
                elif operation_check and overwrite:
                    mainlog.info(
                        f"[{flag}@{table}] This process has already been run, but overwrite=True so it has been permitted."
                    )
                else:
                    pass

                engine = self._sql_engine

                _base_wrapper(*(engine, table, *args), **kwargs)
                self.meta_add(flag, table)

            return wrapper

        return _chunk_db_table_operation

    @staticmethod
    def _args_gen_OBJECT_TYPES(
        instance: Any, process_flag: str, table: str
    ) -> tuple[list[Any], dict]:
        try:
            return [table, instance.schema.table_schema[table]], {}
        except KeyError as err:
            raise ValueError(
                f"[{process_flag}@{table}] Failed to fetch args / kwargs. MSG={err.__str__()}"
            )

    @chunk_db_table_operation(
        "CORRECT_OBJ_TYPES",
        afunc=_args_gen_OBJECT_TYPES,
        allow_overwrite=False,
        inplace=True,
    )
    def correct_object_types(self, table, *args, **kwargs):
        """
        Standardized process to correct the object type in a cross-matching table.

        Generically, when a :py:class:`CrossMatchDatabase` is generated, the match_tables are not corrected for replace their
        object types with the SIMBAD standard that ``pyXMIP`` recognizes. This process is designed to perform the conversion.

        Parameters
        ----------
        tables: list[str], optional
            The tables to perform the reduction process on. If this parameter is not specified, then all of the tables will
            be processed.
        overwrite: bool, optional
            If ``True``, then the process will be performed regardless of whether or not it has been performed before.

            .. warning::

                This will cause the process to be performed on a table which may already have the process run. This will
                often be either redundant or break your code. Use with caution.
        chunksize: int, optional
            The size of the chunks to pass each table through in. In effect, this controls how memory intensive the
            process is. For each chunk, the table is read in, processed, and written to a temp table in the SQL file. As such,
            smaller chunksizes will lead to longer read / write times but lower memory usage. By default, ``chunksize = 10000``.

        """
        _ = kwargs
        converted_chunk = SourceTable._convert_types(table, args[1])
        return converted_chunk

    @chunk_db_table_operation(
        "STNDIZE_COORDS",
        afunc=_args_gen_OBJECT_TYPES,
        allow_overwrite=False,
        inplace=True,
    )
    def standardize_coordinates(self, table, *args, **kwargs):
        """
        Ensure that all cross-matching tables provided (``tables``) have an RA and DEC column in them. This can then
        be used as the standard backbone coordinate system for all reductions.

        Parameters
        ----------
        tables: list[str], optional
            The tables to perform the reduction process on. If this parameter is not specified, then all of the tables will
            be processed.
        overwrite: bool, optional
            If ``True``, then the process will be performed regardless of whether or not it has been performed before.

            .. warning::

                This will cause the process to be performed on a table which may already have the process run. This will
                often be either redundant or break your code. Use with caution.
        chunksize: int, optional
            The size of the chunks to pass each table through in. In effect, this controls how memory intensive the
            process is. For each chunk, the table is read in, processed, and written to a temp table in the SQL file. As such,
            smaller chunksizes will lead to longer read / write times but lower memory usage. By default, ``chunksize = 10000``.

        """
        table_name, schema = args

        coordinates = SkyCoord(
            *[
                np.array(table[col.name]) * col.unit
                for col in schema.coordinate_columns
            ],
            frame=schema.coordinate_frame,
        )
        icrs_coords = coordinates.transform_to(ICRS)
        table["RA"], table["DEC"] = (
            icrs_coords.frame.spherical.lon.deg,
            icrs_coords.frame.spherical.lat.deg,
        )

        return table

    @chunk_db_table_operation(
        "SCORE",
        afunc=lambda instance, table, flag: ([], {}),
        allow_overwrite=True,
        inplace=True,
    )
    def _score(self, table_chunk, *args, **kwargs):
        _ = kwargs
        weight_dict = args[0]
        table_chunk["SCORE"] = np.sum(
            np.array(
                [
                    v * np.array(table_chunk[f"{k}_SCORE"])
                    for k, v in weight_dict.items()
                ]
            ),
            axis=0,
        )

        return table_chunk

    def add_catalog(self, catalog_path, schema=None, overwrite=False, **kwargs):
        r"""
        Add the catalog to the database.

        Parameters
        ----------
        catalog_path: str
            The path to the catalog data file.
        schema: :py:class:`schema.SourceTableSchema`, optional
            The schema to associate with the catalog. If a schema is not provided then the system will attempt to construct
            one.
        overwrite: bool, optional
            If ``True``, the process will run regardless of whether or not there is an existing CATALOG table.
        kwargs:
            Additional arguments to pass through the method.

        Returns
        -------
        None
        """
        # ========================================= #
        # Setting up the procedure                  #
        # ========================================= #

        # -- looking for the catalog path -- #
        catalog_path = pt.Path(catalog_path)
        assert catalog_path.exists(), f"The catalog at {catalog_path} doesn't exist."

        mainlog.debug(f"Adding {catalog_path} to {self.path}.")
        # Load the catalog into memory.
        catalog = SourceTable.read(catalog_path, format=kwargs.pop("format", None))

        if schema is not None:
            catalog.schema = schema

        self.add_catalog_from_table(catalog, overwrite=overwrite, **kwargs)

    def add_catalog_from_table(self, catalog, overwrite=False, **kwargs):
        """
        Add the catalog to the database.

        Parameters
        ----------
        catalog: :py:class:`structures.table.SourceTable`
            The source table to add.
        overwrite: bool, optional
            If ``True``, the process will run regardless of whether or not there is an existing CATALOG table.
        kwargs:
            Additional arguments to pass through the method.

        Returns
        -------
        None
        """

        _pname = "CATALOG_INCLUDED"
        # -- managing the schema -- #
        schema = catalog.schema

        # We need to assure that we have a NAME column.
        assert (
            schema.NAME is not None
        ), "The schema doesn't have a directive for the NAME column. Try manually providing a schema."
        mainlog.debug(
            f"Schema indicates {schema.NAME} is the object identifier. Renaming to CATOBJ."
        )
        # Coercing format
        # We attach the catalog as-is except for the name column.
        # The name column gets renamed to CATALOG_OBJECT

        catalog.rename_column(schema.NAME, "CATOBJ")
        catalog.remove_columns(kwargs.pop("ignore_columns", []))

        # ========================================= #
        # Run the procedure                         #
        # ========================================= #
        if self.has_catalog:
            if overwrite:
                mainlog.warning(
                    f"Process {_pname} has already been completed. Overwriting..."
                )
            else:
                mainlog.error(
                    f"Process {_pname} has already been completed. Skipping..."
                )
                return None

        # -- fixing datatypes -- #
        # This must be done because strings might come back as bytestrings -> np.objects -> BLOB in sql.
        catalog = catalog.to_pandas()

        for col, dtype in catalog.dtypes.items():
            if dtype == object:
                catalog[col] = catalog.loc[:, col].astype("string")

        # -- passing off to write -- #
        with self._sql_engine.connect() as conn:
            catalog.to_sql("CATALOG", conn, index=False, if_exists="replace")

        self.meta_add(_pname, "all")

    def _run_basic_corrections(self, catalog, tables=None, overwrite=False, **kwargs):
        """
        Run the basic baseline corrections.

        .. warning::

            This method is a standard part of the post-processing procedure. It is unlikely you would
            ever need to run this manually. Proceed with caution!

        Parameters
        ----------
        catalog: :py:class:`structures.table.SourceTable`
            The catalog used for the cross-matching.
        tables: list of str, optional
            The tables to perform this correction on. If unspecified, then all tables are corrected.
        registry: :py:class:`structures.databases.DBRegistry`, optional
            A database registry in which all of the listed databases are found. By default, this will be the
            built-in database registry.
        overwrite: bool, optional
            If ``True``, this will allow the process to run even if it has already been performed.

            .. warning::

                In most cases, this is unadvised because it will most likely fail.

        Returns
        -------
        None
        """
        # --------------------------------------- #
        # Setup processes and determine tables    #
        # --------------------------------------- #
        if tables is None:
            tables = self.match_tables

        # -- Adding the catalog to table -- #
        self.add_catalog_from_table(catalog, overwrite=overwrite, **kwargs)

        # -- Performing by-table corrections -- #
        for table in tables:
            self.correct_object_types(table, overwrite=overwrite, **kwargs)
            self.standardize_coordinates(table, overwrite=overwrite, **kwargs)

    def run_reduction(
        self, reduction_process: RProc, table: str, overwrite: bool = False
    ):
        """
        Run the provided reduction process on the specified table of this cross-matching database.

        Parameters
        ----------
        reduction_process: :py:class:`structures.reduction.ReductionProcess`
            The reduction process to perform on the specified ``table``.
        table: str
            The name of the table on which to perform the reduction process.
        overwrite: bool
            If the reduction process has already been performed on this table, then ``True`` will allow it to proceed while
            ``False`` will cause it to fail.

        """
        # -- check against the meta table -- #
        if self.check_meta(reduction_process.process_name, table):
            # The process has been performed before.
            if not overwrite:
                raise ValueError(
                    f"Reduction process {reduction_process} was already performed on {table} and overwrite = False."
                )
            else:
                mainlog.warning(
                    f"Reduction process {reduction_process} was already performed on {table}. Overwriting results..."
                )
                self.meta_remove(reduction_process.process_name, table)
                # remove from meta so we don't get duplicates later.

        # -- Run the reduction -- #
        reduction_process(self, table)  # --> runs the reduction.

        # -- post-processing -- #
        self.meta_add(
            reduction_process.process_name, table
        )  # Add the new reduction process to the table.

    def score_matches(
        self,
        weight: dict[tuple | str, float],
        tables: Collection[str] | None = None,
        show_summary: bool = False,
    ):
        r"""
        Score the :py:class:`cross_reference.CrossMatchDatabase` instance.

        Parameters
        ----------
        weight: dict of tuple: float or str: float
            The scoring weights for each of the reduction processes. Keys correspond to each of the reduction processes and
            the values correspond to process weight (:math:`\alpha_i \in [0,1]`). If keys are ``str``, then the weight is applied
            universally across tables. If the keys are ``tuple`` of the form ``('<reduction_process>','<table>')`` then the weights
            will be applied differently to different tables.
        tables: list of str, optional
            The tables to perform the scoring procedure on.

            This kwarg is only applied to weights specified with only the reduction process name. If weights are specified
            with both reduction process and table name, then this will overwrite the applied tables.
        show_summary: bool, optional
            If ``True``, then show the summary.
        """
        if not tables:
            tables = self.match_tables
        # ---------------------------------------------- #
        # Enforcing weight restrictions                  #
        # ---------------------------------------------- #
        # restructure the weight dictionary to contain tuples specifically.
        _altered_weights = {}

        for k, v in weight.items():
            if isinstance(k, str):
                _altered_weights = {
                    **_altered_weights,
                    **{(k, _sk): v for _sk in tables},
                }
            else:
                _altered_weights = {**_altered_weights, **{k: v}}

        weights = _altered_weights

        # re-weighting the weights.
        _total_tables = list(set([k[1] for k in _altered_weights]))

        for _tbl in _total_tables:
            _subweight_dict = {
                k: v for k, v in _altered_weights.items() if k[1] == _tbl
            }
            _weight_sum = np.sum(
                list(_subweight_dict.values())
            )  # compute the total weight value.
            # If the weights are not weighted correctly, we need to re-weight. *1/(sum of weights).
            for k, _ in _subweight_dict.items():
                weights[k] *= 1 / _weight_sum

        # ---------------------------------------------- #
        # Debugging summary                              #
        # ---------------------------------------------- #
        if show_summary:
            mainlog.debug(f"SCORING WEIGHT SUMMARY: {self}")
            for table_name in _total_tables:
                _subweights = {k: v for k, v in weights.items() if k[1] == table_name}
                mainlog.debug(f"\t[{table_name}] N reductions = {len(_subweights)}")

                for _k, _v in _subweights.items():
                    mainlog.debug(f"\t\t[{_k}] Weight = {_v}")

        # ----------------------------------------------- #
        # Performing the computations                     #
        # ----------------------------------------------- #
        # This should then be performed in chunked operations on each table.
        for table in tqdm(_total_tables):
            # cycle through all of the tables in the listing.

            # -- Construct the weight dictionary -- #
            _subweights = {k[0]: v for k, v in weights.items() if k[1] == table}

            self._score(table, _subweights)

    def plot_matches(
        self,
        catalog_object,
        table,
        fig=None,
        cmap=None,
        norm=None,
        vmin=None,
        vmax=None,
        resolution=300,
        fov="5 arcmin",
        hips_kwargs=None,
        scatter_kwargs=None,
        **kwargs,
    ):
        """
        Plot the matches to a given catalog source.

        Parameters
        ----------
        catalog_object: str
            The ``CATOBJ`` identifier for the object.
        table: str
            The table to search for matches in.
        fig: :py:class:`plt.axes.Figure`, optional
            The figure to add the axes to.
        cmap: str or :py:class:`plt.cm.ColorMap`
            The colormap to use. Default is viridis.
        norm: :py:class:`plt.colors.Norm`
            The norm to use for the colormap.
        vmin: float
            The minimum value of the background image.
        vmax: float
            The maximum value of the background image.
        resolution: int
            The number of pixels per side of the image (when using HIPS).
        fov: :py:class:`astropy.units.Quantity`
            The field of view for the HIPS image.
        hips_kwargs: dict
            Additional kwargs to pass through the hips generation system.

            +--------------------+-----------+---------------------------------------------------------------+
            | Name               | Type      | Description                                                   |
            +====================+===========+===============================================================+
            | ``enabled``        | `bool`    | [default: ``True``] Use a HIPS image as the background?       |
            +--------------------+-----------+---------------------------------------------------------------+


        scatter_kwargs: dict
            Additional kwargs to pass through scatter.
        **kwargs
            Additional key-word arguments as listed below.

            +--------------------+-----------+---------------------------------------------------------------+
            | Name               | Type      | Description                                                   |
            +====================+===========+===============================================================+
            | ``figsize``        | `tuple`   | [default: ``(10,10)``] The size of the figure to generate.    |
            +--------------------+-----------+---------------------------------------------------------------+


        Returns
        -------

        """
        import matplotlib.pyplot as plt
        from astropy.coordinates import SkyCoord
        from astropy.wcs import WCS

        from pyXMIP.utilities.plot import get_hips_image

        # ============================================================ #
        # Setting up runtime variables                                 #
        # ============================================================ #
        # -- kwargs / args management -- #
        hips_kwargs, scatter_kwargs = (
            {} if hips_kwargs is None else hips_kwargs,
            {} if scatter_kwargs is None else scatter_kwargs,
        )

        if fig is None:
            fig = plt.figure(figsize=kwargs.pop("figsize", (10, 10)))

        # -- Setup central object parameters -- #
        assert (
            self.has_catalog
        ), "Cannot plot matches without a loaded CATALOG table. Try adding the catalog."

        # -- Manage the central object information -- #
        object_data = self.query(
            f"SELECT * FROM CATALOG WHERE CATALOG_OBJECT == '{catalog_object}'"
        )
        RAC, DECC = object_data["RA"], object_data["DEC"]
        central_position = SkyCoord(ra=RAC, dec=DECC, unit="deg")

        # ============================================================ #
        # Pull object positions                                        #
        # ============================================================ #
        data_table = self.query(
            f"SELECT * FROM {table} WHERE CATOBJ == '{catalog_object}'"
        )

        # -- pull data -- #
        _ra, _dec = data_table["RA"], data_table["DEC"]

        # ============================================================ #
        # Producing the Image                                          #
        # ============================================================ #

        if hips_kwargs.pop("enabled", True):
            # We are using a HIPs map.
            hips_image = get_hips_image(
                central_position, fov, (resolution, resolution), **hips_kwargs
            )
            wcs = WCS(hips_image[0].header)
            ax = fig.add_subplot(111, projection=wcs)
            ax.imshow(hips_image[0].data, cmap=cmap, norm=norm, vmin=vmin, vmax=vmax)
        else:
            from astropy.coordinates import Angle

            # We are not using a HIPs map, we need to perform this by hand.
            _fov = Angle(fov)

            _custom_wcs = {
                "CTYPE1": "RA---TAN",
                "CTYPE2": "DEC---TAN",
                "CUNIT1": "deg",
                "CUNIT2": "deg",
                "NAXIS1": resolution,
                "NAXIS2": resolution,
                "CRPIX1": resolution // 2,
                "CRPIX2": resolution // 2,
                "CRVAL1": RAC,
                "CRVAL2": DECC,
                "CDELT1": (1.2 * (2 * _fov) / resolution).to_value("deg"),
                "CDELT2": (1.2 * (2 * _fov) / resolution).to_value("deg"),
            }

            wcs = WCS(_custom_wcs)
            ax = fig.add_subplot(111, projection=wcs)

        # -- Manage the scatter plot -- #

        # Fetch the corrected parameters
        for _sra, _sdec in zip(_ra, _dec):
            ax.scatter(
                _sra,
                _sdec,
                transform=ax.get_transform("world"),
                **scatter_kwargs,
            )

        # adding source position scatter
        ax.scatter(
            central_position.ra,
            central_position.dec,
            transform=ax.get_transform("world"),
            s=(72 * fig.get_size_inches()[0] / 50) ** 2,
            color="red",
            marker="+",
        )
        # ============================================================ #
        # Returning outputs                                            #
        # ============================================================ #
        return ax, fig

    def connect(self):
        return self._sql_engine.connect()

    @classmethod
    def from_file(cls, path):
        """
        Load a :py:class:`CrossMatchDatabase` from file.

        Parameters
        ----------
        path: str
            The path from which to load the database.

        Returns
        -------
        :py:class:`CrossMatchDatabase`
            The returned matching database.

        """
        return cls(path)


PydanticCMD = Annotated[CrossMatchDatabase, _CMDTypePydanticAnnotation]


# ============================================================================================ #
# X-Matching Processes                                                                         #
# ============================================================================================ #
def cross_match(
    input_path: str | pt.Path,
    output_path: str | pt.Path,
    databases: list[SourceDatabase | str] = None,
    registry: DBRegistry = None,
    overwrite: bool = False,
    *args,
    **kwargs,
) -> CrossMatchDatabase:
    r"""
    Cross match a table of known objects against a set of databases and output the result to a path of your choice.

    Parameters
    ----------
    input_path: str
        The path to the input file (must be a table with a readable file format by ``astropy``).

        .. hint::

            This should be the catalog of sources which is to be cross-matched against the databases

    output_path: str, Path
        The ``.db`` file to write the output to. This will be a ``sqlite`` database containing all of the cross-matching
        output data.
    databases: list of :py:class:`structures.databases.SourceDatabase` or list of str, optional
        The databases to cross-match the catalog against. By default, this will be all of the databases currently loaded
        in the ``registry`` :py:class:`structures.databases.DBRegistry` instance provided.

        If entries are ``str``, then they will be looked up in the registry. If they are not, they will be taken as is.

    registry: :py:class:`structures.database.DBRegistry`
        The database registry to lookup databases from. If not provided, defaults to :py:attr:`structures.databases.DEFAULT_DATABASE_REGISTRY`.
    overwrite: bool, optional
        If ``True``, you will be allowed to overwrite a pre-existing ``.db`` file.

    Returns
    -------
    CrossMatchDatabase
        The output matching database.

    See Also
    --------
    :py:func:`cross_match_table`
    """
    mainlog.info(f"X-Matching {input_path} into {output_path}.")

    # ======================================================== #
    # Managing arguments and kwargs. Enforcing types.          #
    # ======================================================== #
    input_path, output_path = pt.Path(input_path), pt.Path(output_path)

    # check the source table reading.
    try:
        source_table = SourceTable.read(input_path)
    except Exception as excep:
        raise ValueError(
            f"Failed to read source table {input_path} because if error {excep.__str__()}."
        )

    # ======================================================== #
    # Running                                                  #
    # ======================================================== #
    return cross_match_table(
        source_table,
        output_path,
        databases=databases,
        registry=registry,
        overwrite=overwrite,
        *args,
        **kwargs,
    )


def cross_match_table(
    table: SourceTable,
    output_path: str | pt.Path,
    databases: list[SourceDatabase | str] = None,
    registry: DBRegistry = None,
    overwrite: bool = False,
    *args,
    **kwargs,
) -> CrossMatchDatabase:
    r"""
    Cross match a table of known objects against a set of databases and output the result to a path of your choice.

    Parameters
    ----------
    table: :py:class:`structures.table.SourceTable`
        The catalog (loaded into memory in a table) to cross-match against databases.
    output_path: str, Path
        The ``.db`` file to write the output to. This will be a ``sqlite`` database containing all of the cross-matching
        output data.
    databases: list of :py:class:`structures.databases.SourceDatabase` or str, optional
        The databases to cross-match the catalog against. By default, this will be all of the databases currently loaded
        in the ``registry`` :py:class:`structures.databases.DBRegistry` instance provided.

        If entries are ``str``, then they will be looked up in the registry. If they are not, they will be taken as is.

    registry: :py:class:`structures.database.DBRegistry`
        The database registry to lookup databases from. If not provided, defaults to :py:attr:`structures.databases.DEFAULT_DATABASE_REGISTRY`.
    overwrite: bool, optional
        If ``True``, you will be allowed to overwrite a pre-existing ``.db`` file.

    Returns
    -------
    CrossMatchDatabase
        The output matching database.

    See Also
    --------
    :py:func:`cross_match`
    """
    import sqlalchemy as sql

    # --------------------------------------- #
    # SETUP
    # --------------------------------------- #
    output_path = pt.Path(output_path)

    # configure the registry
    if registry is None:
        # we default to the standard database registry.
        registry = DEFAULT_DATABASE_REGISTRY

    # configure the databases
    if databases is None:
        databases = list(registry.values())
    else:
        # we need to check the databases.
        databases = [
            database if isinstance(database, SourceDatabase) else registry[database]
            for database in databases
        ]

    mainlog.info(
        f"Cross matching with {len(databases)} databases: {[db.name for db in databases]}."
    )

    # ===================================================== #
    # Setting up the SQL                                    #
    # ===================================================== #
    if output_path.exists():
        # check if the specific table is there
        _tmp_engine = sql.create_engine(f"sqlite:///{output_path}")
        _insp = sql.inspect(_tmp_engine)
        table_names = _insp.get_table_names()

        if "META" in table_names and overwrite:
            mainlog.warning(f"Table META of {output_path} exists. Deleting.")
            with _tmp_engine.connect() as conn:
                conn.execute(sql.text("DROP TABLE META"))

        for tbl in table_names:
            if tbl in [f"{db.name}_MATCH" for db in databases]:
                if not overwrite:
                    raise ValueError(
                        f"Table {tbl} exists in {output_path} and overwrite=False."
                    )
                else:
                    mainlog.warning(
                        f"Table {tbl} exists in {output_path}. Overwrite = True -> deleting."
                    )

                    with _tmp_engine.connect() as conn:
                        _exec = sql.text(f"DROP TABLE '{tbl}'")
                        conn.execute(_exec)

    # ===================================================== #
    # Running                                               #
    # ===================================================== #
    _schema_generator = {"table_schema": table.schema, "db_schema": {}}

    with logging_redirect_tqdm(loggers=[mainlog]):
        # noinspection PyTypeChecker
        for database in tqdm(databases, desc="Cross-Matching"):
            database.source_match(output_path, table, *args, **kwargs)

            # We need to add the database query schema to the schemas collection.
            _table_name = f"{database.name}_MATCH"
            _schema_generator["db_schema"][_table_name] = database.query_schema

    # ===================================================== #
    # Return                                                #
    # ===================================================== #
    mainlog.info("Post-processing the cross-matching database.")

    # -- create the CMDschema -- #
    schema = CMDSchema(
        table_schema=_schema_generator["db_schema"],
        catalog_schema=_schema_generator["table_schema"],
    )

    cmd = CrossMatchDatabase(output_path, schema=schema, overwrite_schema=True)
    cmd._run_basic_corrections(table)

    return CrossMatchDatabase(output_path)



"""
Testing module for pyXMIP's catalogs.
"""
import os
import pathlib as pt

import pytest

from pyXMIP import load
from pyXMIP.utilities.core import bin_directory

test_catalog_path = pt.Path(os.path.join(bin_directory, "testobj", "test_catalog.fits"))


class TestCatalog:
    """
    Test the base catalog.
    """

    @pytest.fixture(scope="class")
    def base_catalog(self):
        yield load(test_catalog_path)

    def test_catalog_exists(self, base_catalog):
        """
        Test that the sample catalog does exist.
        """
        assert 1, "The sample catalog loaded successfully."

    def test_schema(self, base_catalog):
        """
        Test that a schema exists and has the information we anticipate.
        """
        _expected_column_map = {
            "RA": {"name": "RA", "unit": "deg"},
            "DEC": {"name": "DEC", "unit": "deg"},
            "RA_ERR": None,
            "DEC_ERR": None,
            "GAL_L": {"name": "LII", "unit": "deg"},
            "GAL_B": {"name": "BII", "unit": "deg"},
            "GAL_L_ERR": None,
            "GAL_B_ERR": None,
            "SEP": None,
            "NAME": {"name": "IAUNAME", "unit": None},
            "Z": None,
            "TYPE": None,
        }

        assert (
            _expected_column_map == base_catalog.schema.column_map.model_dump()
        ), "The test catalog has a different column map in its schema than expected."

    def test_coordinates(self, base_catalog):
        """
        Simply check that RA and DEC are the same as we expect them to be.
        """
        from numpy.testing import assert_allclose

        lat, lon = base_catalog.lat.to_value("deg"), base_catalog.lon.to_value("deg")
        _elat = [
            -31.0361403,
            -28.53148995,
            -29.51964773,
            -31.20848839,
            -21.6990906,
            -23.18313657,
            -19.9442472,
            -18.54302228,
            -16.23896735,
            -16.81838409,
        ]
        _elon = [
            7.40528954,
            10.52964788,
            12.34260232,
            13.7041894,
            22.29515815,
            20.90908527,
            24.37178961,
            28.07930583,
            29.33960449,
            30.15507953,
        ]

        assert_allclose(_elat, lat, rtol=1e-3)
        assert_allclose(_elon, lon, rtol=1e-3)



"""
Command's specific to the CLI interface.
"""
from argparse import ArgumentParser
from typing import Mapping

from pyXMIP.utilities.text import get_package_version, print_version

CLI_FUNCTIONS = {f.__name__: f for f in [get_package_version, print_version]}


class CLIParser(ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def build_recursive(cls, parser_config: Mapping, base_parser=None):
        if base_parser is None:
            base_parser = cls(prog="pyXMIP")

            _w = base_parser.add_subparsers(title="commands", help="Available commands")

            cls.build_recursive(parser_config, base_parser=_w)

            return base_parser

        # -- Iterate through the layers -- #
        for k, v in parser_config.items():
            _is_command = (
                "subparsers" not in v.keys()
            )  # determines if its a command or not.

            if _is_command:
                # This entry is a command.
                _s = base_parser.add_parser(name=k, help=v.get("help", None))

                for arg_name, arg_dict in v.get("args", {}).items():
                    _s.add_argument(name=arg_name, **arg_dict)
                for kwarg_name, kwarg_dict in v.get("kwargs", {}).items():
                    _s.add_argument(name=kwarg_name, **kwarg_dict)

                _s.set_defaults(operation=v.get("function", "none"))

            else:
                _s = base_parser.add_parser(name=k, help=v.get("help", None))
                _q = _s.add_subparsers(title=v.get("title"), help=v.get("help", None))

                cls.build_recursive(v, base_parser=_q)

        return base_parser




"""
Regression tools for pyXMIP
"""
from types import SimpleNamespace

import numpy as np
from astropy.coordinates import SkyCoord
from sklearn.model_selection import train_test_split
from sklearn.neighbors import KNeighborsRegressor, RadiusNeighborsRegressor
from tqdm.contrib.concurrent import process_map

from pyXMIP.utilities.geo import convert_skycoord
from pyXMIP.utilities.logging import mainlog


class PoissonMapRegressor:
    """
    Generic wrapper class for :py:mod:`sklearn` regression classes for generating :py:class:`pyXMIP.structures.map.Map` instances.

    Attributes
    ----------

    regressor_parameters: dict
        The complete set of parameters for this regressor class. This includes hyperparameters and also
        parameters which are not permitted to vary.
    default_training_fraction: float
        The standard training fraction for the regressor.
    hyperparameters: dict
        Subset of ``regressor_parameters`` containing only the (variation permissible) hyperparameters of the regressor.
    """

    regressor_parameters = {}
    default_training_fraction = 0.5
    hyperparameters = {}

    def __init__(self, regressor_class, **kwargs):
        """
        Initialize the :py:class:`PoissonMapRegressor` instance.

        Parameters
        ----------
        regressor_class:
            The :py:mod:`sklearn` specific regressor type to build this class around.
        **kwargs
            regression parameters.
        """
        self.regressor = regressor_class(
            **{
                k: (v if k not in kwargs else kwargs[k])
                for k, v in self.__class__.regressor_parameters.items()
            }
        )

    def train_model(self, count_table, object_type, training_fraction=None, **kwargs):
        """
        Train the regression model on a ``COUNTS`` table.

        Parameters
        ----------
        count_table: :py:class:`astropy.table.Table`
            The table of counts to use for the training.
        object_type: str
            The specific object type to model.
        training_fraction: float
            The training fraction.
        **kwargs
            Additional kwargs.
        """
        # -- Setup and sanity check -- #
        training_fraction = (
            training_fraction
            if training_fraction is not None
            else self.__class__.default_training_fraction
        )
        assert (
            object_type in count_table.columns
        ), f"Failed to find {object_type} in provided table."

        # -- Fetching data -- #
        X, Y = self._sanitize_points(count_table, object_type)
        mainlog.debug(
            f"Training {self.regressor} on {int(training_fraction * Y.size)} data-points"
        )

        positions_train, positions_test, values_train, values_test = train_test_split(
            X,
            Y,
            train_size=training_fraction,
            random_state=kwargs.pop("random_state", None),
        )
        self.regressor.fit(positions_train, values_train)

        if not kwargs.pop("return_all", False):
            return SimpleNamespace(
                score=self.regressor.score(positions_test, values_test),
                x_train=None,
                x_test=None,
                y_train=None,
                y_test=None,
            )
        else:
            return SimpleNamespace(
                score=self.regressor.score(positions_test, values_test),
                x_train=positions_train,
                x_test=positions_test,
                y_train=values_train,
                y_test=values_test,
            )

    def cross_validate(
        self, count_table, object_type, training_kw=None, param_kw=None, **kwargs
    ):
        """
        Cross-validate the regressor against the available count data.

        Parameters
        ----------
        count_table: :py:class:`astropy.table.Table`
            The table of data from ``COUNTS``.
        object_type: str
            The object type to pull from the table.
        training_kw: dict
            Training kwargs.
        param_kw: dict
            Parameter kwargs
        kwargs:
            Additional kwargs to pass.
        """
        from sklearn.model_selection import GridSearchCV

        mainlog.debug(f"Cross-validating {self}.")
        # -- SETUP -- #
        # grab data
        X, Y = self._sanitize_points(count_table, object_type)

        # fix parameters
        if training_kw is None:
            training_kw = {
                "training_fraction": self.__class__.default_training_fraction,
                "cv": 5,
            }

        if param_kw is None:
            param_kw = self.__class__.hyperparameters

        # split data
        positions_train, positions_test, values_train, values_test = train_test_split(
            X, Y, train_size=training_kw.get("training_fraction")
        )

        # -- RUN GS -- #
        cv_estimator = GridSearchCV(self.regressor, param_kw, **kwargs)
        cv_estimator.fit(positions_train, values_train)

        return cv_estimator.cv_results_

    def _sanitize_points(self, count_table, object_type):
        """Pull count data and fix issues."""
        RA, DEC, N, R = (
            np.array(count_table["RA"]),
            np.array(count_table["DEC"]),
            np.array(count_table[object_type]),
            np.array(count_table["RAD"]),
        )

        # RA and DEC need to be converted to standardized lat-lon. USE MATH CONVENTION
        positions = SkyCoord(ra=RA, dec=DEC, unit="deg")
        phi, theta = convert_skycoord(positions, "math_convention")
        # phi,theta = np.deg2rad(RA),np.deg2rad(DEC)
        # Fixing R and getting areas.
        # NOTE: we always use N/arcmin^2. The RAD is always already in ARCMIN.
        A = np.pi * (R**2)
        density = N / A

        X = np.vstack([theta, phi]).T  # --> Enforced haversine convention.
        Y = density

        return X, Y


class BayesianPoissonMapRegressor:
    r"""
    The :py:class:`BaysianPoissonMapRegressor` class provides a variety of posterior estimation methods for estimating
    sky-map densities.

    Notes
    -----

    The :py:class:`BaysianPoissonMapRegressor` takes 2 kwargs: ``prior`` and ``model``. Generically, the posterior distribution is

    .. math::

        P(\beta|f,\textbf{r},\textbf{N},\textbf{A}) = \frac{p(\beta)}{p(\textbf{N}|\textbf{A},\textbf{r},\beta)} \prod_{i=1}^{i = |N|} \frac{(A_if(\textbf{r_i},\beta))^{N_i} \exp\left(-A_if(\textbf{r_i},\beta)\right)}{N_i!}

    Thus, the log posterior is

    .. math::

        \log P(\beta|f,\textbf{r},\textbf{N},\textbf{A}) = p(\beta) + \sum_{i=1}^{i=|N|} N_i \log(A_if(\textbf{r_i},\beta)) + A_if(\textbf{r_i},\beta) - N_i! + C.

    If no model is specified (``model=None``), then the posterior instead takes the form

    .. math::

        \log P(\lambda_k|f,\textbf{r}_k,\textbf{N}_k,\textbf{A}_k) = p(\lambda_k) + \sum_{i=1}^{i=|N_k|} N_i \log(A_if(\textbf{r_i},\beta)) + A_if(\textbf{r_i},\beta) - N_i! + C.

    Where :math:`\lambda_k` is the rate in each of the HEALPix cells. Thus, when no model is specified, the problem is discretized between cells.
    """

    def __init__(self, prior=None, model=None):
        self.model = model
        self.prior = prior

    def build_map_MAP(self, count_table, object_type, **kwargs):
        # ======================================= #
        # Setup
        # ======================================= #
        mainlog.debug(f"Building MAP estimate for {self}.")

        # ======================================= #
        # organize the pipeline, select method
        # ======================================= #
        if (self.model is None) and (self.prior is None):
            # We can get away with an analytical discretization
            mainlog.debug("MAP estimator -- ANALYTIC [no p, no m]")
            return self._build_map_MAP_analytic(count_table, object_type, **kwargs)
        else:
            raise NotImplementedError(
                "Cannot currently generate MAP estimates when priors / model are used."
            )

    def _build_map_MAP_analytic(self, count_table, object_type, **kwargs):
        _ = kwargs
        count_table = count_table
        # ==================================== #
        # Discretizing data observations
        # ==================================== #
        PIX_ID, N, RAD = (
            count_table["PIX_ID"],
            count_table[object_type],
            count_table["RAD"],
        )
        A = np.array(np.pi * (RAD**2), dtype="float64")
        PIX_ID = np.array(PIX_ID, dtype="uint32")
        N = np.array(N, dtype="uint32")

        # ==================================== #
        # Getting Estimates
        # ==================================== #
        from itertools import repeat

        from pyXMIP.utilities.optimize import created_shared_memory_equivalent, split

        estimates = np.zeros(
            np.amax(PIX_ID) + 1, dtype="float64"
        )  # generate PIX_ID count. ! Not necessarily all HP, might cut off end.
        mp_kwargs = kwargs.pop(
            "mp_kwargs", {"multiprocess": True, "nproc": 8, "chunksize": 100}
        )
        if not mp_kwargs.get("multiprocess", True):
            mp_kwargs = {"multiprocess": False, "nproc": 1, "chunksize": len(estimates)}
        # Produce the shared memory regions #
        SMI, SMN, SMA, SME = (
            created_shared_memory_equivalent(PIX_ID),
            created_shared_memory_equivalent(N),
            created_shared_memory_equivalent(A),
            created_shared_memory_equivalent(estimates),
        )
        # Chunk generation #
        PIX_GROUPS = split(PIX_ID, len(PIX_ID) // mp_kwargs.get("chunksize", 100))
        # run #
        process_map(
            self._get_map_analytic_estimates_mp,
            PIX_GROUPS,
            repeat(SMI),
            repeat(SMN),
            repeat(SMA),
            repeat(SME),
            max_workers=mp_kwargs.get("nproc", 8),
            desc="Computing MAP estimate (ANYTC)",
        )

        # recover shared memory regions #
        ee = np.frombuffer(SME.buf)[:]
        estimates = ee.copy()
        del ee
        for u in [SMI, SMN, SMA, SME]:
            u.close()
            u.unlink()
        return estimates

    @staticmethod
    def _get_map_analytic_estimates_mp(pxg, PIX_ID, N, A, E):
        n, a, ee, i = (
            np.frombuffer(N.buf, dtype="uint32"),
            np.frombuffer(A.buf),
            np.frombuffer(E.buf),
            np.frombuffer(PIX_ID.buf, dtype="uint32"),
        )

        for pi in pxg:
            if pi in i:
                ee[pi] = np.sum(n[np.where(i == pi)]) / np.sum(a[np.where(i == pi)])


class KNNeighborMapRegressor(PoissonMapRegressor):
    regression_class = KNeighborsRegressor
    regressor_parameters = {
        "n_neighbors": 10,
        "weights": "distance",
        "algorithm": "auto",
        "metric": "haversine",
    }
    hyperparameters = {"n_neighbors": np.array([5, 10, 30, 60, 120, 300])}

    def __init__(self, **kwargs):
        super().__init__(self.__class__.regression_class, **kwargs)


class RNNeighborMapRegressor(PoissonMapRegressor):
    regression_class = RadiusNeighborsRegressor
    regressor_parameters = {
        "radius": 0.01,
        "weights": "distance",
        "algorithm": "auto",
        "metric": "haversine",
    }
    hyperparameters = {"radius": np.array([0.001, 0.005, 0.01, 0.05, 0.1])}

    def __init__(self, **kwargs):
        super().__init__(self.__class__.regression_class, **kwargs)


def train_map_regressor(
    regressor, map_phi, map_theta, map_target_values, training_fraction=0.5, **kwargs
):
    """
    Train the map regressor ``regressor`` on the provided positional data.

    Parameters
    ----------
    regressor
    map_phi
    map_theta
    map_target_values
    kwargs

    Returns
    -------

    """
    mainlog.debug(
        f"Training {regressor} on {int(training_fraction*len(map_target_values))} data-points"
    )
    positions = np.vstack(map_phi, map_theta).T  # ready for sklearn now.
    positions_train, positions_test, values_train, values_test = train_test_split(
        positions,
        map_target_values,
        train_size=training_fraction,
        random_state=kwargs.pop("random_state", None),
    )

    regressor.fit(positions_train, values_train)
    return regressor.score(positions_test, values_test)



"""Text utilities for the pyXMIP backend.

Notes
-----

The :py:mod:`utilities.text` module contains various utility functions for text output used elsewhere
in this package.
"""
import os
import pathlib as pt

from pyXMIP.utilities.core import bin_directory

# ------------------------------------------------------------------ #
# Globals                                                            #
# ------------------------------------------------------------------ #
_setup_tools_directory = os.path.join(pt.Path(__file__).parents[2], "setup.py")


# ------------------------------------------------------------------ #
# Simple utility functions                                           #
# ------------------------------------------------------------------ #
def get_package_version() -> str | None:
    """
    Determine the version of the currently installed version.
    """
    try:
        with open(_setup_tools_directory, "r") as setup_file:
            for line in setup_file:
                if line.strip().startswith("version="):
                    # Extract the version string (assuming it's in single or double quotes)
                    version = (
                        line.split("=")[1].strip().strip(",").strip(r"'").strip('"')
                    )
                    return version
    except FileNotFoundError:
        return None


def print_version() -> None:
    print(get_package_version())


def print_cli_header() -> None:
    """
    Print the CLI title.
    """
    with open(os.path.join(bin_directory, "txt", "asciilogo.txt"), "r") as fo:
        header = fo.read()

    print(header % dict(version=get_package_version()))



"""
Utility module for optimized operations on datasets or via parallelism
"""
from multiprocessing.shared_memory import SharedMemory
from typing import Callable, Collection, Iterator, TypeVar

import numpy as np

_T = TypeVar("_T")


def split(a: Collection[_T], n: int) -> Collection[Collection[_T]]:
    """
    Split the collection ``a`` into ``n`` chunks of near-equally sized data.

    Parameters
    ----------
    a: list
        The list to split into chunks.
    n: int
        The number of chunks.

    Returns
    -------
    list of list
        A nested list of each chunk.

    """
    k, m = divmod(len(a), n)
    return [a[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(n)]


def created_shared_memory_equivalent(array: np.ndarray) -> SharedMemory:
    """
    Create a shared memory equivalent of an array. Effectively produces a pointer-like variable directed
    to the position in memory where this array is stored.

    Parameters
    ----------
    array: array
        The array to generate a shared memory equivalent of.

    Returns
    -------
    SharedMemory
        The buffer specification for the data in memory.

    """
    header = SharedMemory(create=True, size=array.nbytes)
    copy = np.ndarray(array.shape, dtype=array.dtype, buffer=header.buf)
    copy[:] = array[:]
    return header


def map_to_threads(
    mappable: Callable[[...], ...], *args, threading_kw: dict = None
) -> Iterator | map:
    """
    Map a function (``mappable``) with arguments ``*args`` to threads or run without threads if threading is
    not enabled.

    Parameters
    ----------
    mappable: Callable
        The mapping function which acts on ``args`` and returns a result.
    *args:
        Arguments to pass through ``mappable``. These would be any of the standard arguments.
    threading_kw: dict
        Dictionary containing the threading parameters. TODO: threading kwargs.

    Returns
    -------
    Iterator
        The output of the mapping operation. Equivalent to ``map(mappable,*args)``.

    Notes
    -----

    This function utilizes the ``concurrent.futures`` ``ThreadPoolExecutor``.
    """

    from concurrent.futures import ThreadPoolExecutor

    # ---------------------------------------- #
    # Setup the threading environment          #
    # ---------------------------------------- #
    # Create the kwargs if they don't exist.
    if threading_kw is None:
        threading_kw = {}

    _max_workers = threading_kw.pop("max_workers", 1)
    if _max_workers == 1:
        _threading = False
    else:
        _threading = True

    # ---------------------------------------- #
    # Run the threads                          #
    # ---------------------------------------- #
    if _threading:
        with ThreadPoolExecutor(
            max_workers=_max_workers,
            thread_name_prefix=threading_kw.pop("thread_name_prefix", ""),
        ) as executor:
            results = executor.map(mappable, *args)
    else:
        results = map(mappable, *args)

    return results



"""
Logging configuration for ``pyXMIP``.

Notes
-----

``pyXMIP`` has the following loggers:

- ``mainlog``: the main logger for the module. This is used for any standard display needs within any
  of the processes.
- ``devlog``: The development logger. Produces logging statements which are more useful for debugging.


"""
import logging
import sys

from pyXMIP.utilities.core import pxconfig

# --------------------------------------------------------- #
# Establishing relevant streams                             #
# --------------------------------------------------------- #
streams = dict(
    mainlog=getattr(sys, pxconfig.config.logging.mainlog.stream),
    devlog=getattr(sys, pxconfig.config.logging.devlog.stream),
)

_loggers = dict(mainlog=logging.Logger("pyXMIP"), devlog=logging.Logger("pyXMIP-DEV"))

_handlers = {}

for k, v in _loggers.items():
    # Construct the formatter string.
    _handlers[k] = logging.StreamHandler(streams[k])
    _handlers[k].setFormatter(
        logging.Formatter(getattr(pxconfig.config.logging, k).format)
    )
    v.addHandler(_handlers[k])
    v.setLevel(getattr(pxconfig.config.logging, k).level)
    v.propagate = False

    if k != "mainlog":
        v.disabled = not getattr(pxconfig.config.logging, k).enabled

mainlog: logging.Logger = _loggers["mainlog"]
# :py:class:`logging.Logger`: The main logger for ``pyXMIP``.
devlog: logging.Logger = _loggers["devlog"]
# :py:class:`logging.Logger`: The development logger for ``pyXMIP``.



"""
Core utilities for many tasks in the ``pyXMIP`` package environment.

Notes
-----

Aside from access to the logging system, this module has very little practical use for the end user.

"""
import functools
import operator
import os
import pathlib as pt
from functools import reduce
from typing import Any, Collection, Iterable, Mapping, Self

import astropy.units as u
import ruamel.yaml
from sqlalchemy.types import TypeEngine

# -- configuration directory -- #
bin_directory: pt.Path = pt.Path(os.path.join(pt.Path(__file__).parents[1], "bin"))
# :py:class:`pathlib.Path`: The directory in which the ``pyXMIP`` bin is located.
config_directory: pt.Path = pt.Path(
    os.path.join(pt.Path(__file__).parents[1], "bin", "config.yaml")
)
# :py:class:`pathlib.Path`: The directory in which the ``pyXMIP`` configuration files are located.

yaml = ruamel.yaml.YAML()
yaml.register_class(u.Quantity)
yaml.register_class(TypeEngine)


class AttrDict(dict):
    """
    Attribute accessible dictionary.
    """

    def __init__(self, mapping: Mapping):
        super(AttrDict, self).__init__(mapping)
        self.__dict__ = self

        for key in self.keys():
            self[key] = self.__class__.from_nested_dict(self[key])

    @classmethod
    def from_nested_dict(cls, data: Any) -> Self:
        """Construct nested AttrDicts from nested dictionaries."""
        if not isinstance(data, dict):
            return data
        else:
            return AttrDict({key: cls.from_nested_dict(data[key]) for key in data})

    @classmethod
    def clean_types(cls, _mapping):
        for k, v in _mapping.items():
            if isinstance(v, AttrDict):
                _mapping[k] = cls.clean_types(_mapping[k])
            else:
                pass
        return dict(_mapping)

    def clean(self):
        return self.clean_types(self)


class YAMLConfiguration:
    """
    General class representing a YAML configuration file.
    """

    def __init__(self, path: pt.Path | str):
        # ------------------------------------------------- #
        # Defining parameters                               #
        # ------------------------------------------------- #
        self.path: pt.Path = pt.Path(path)
        # :py:class:`pathlib.Path`: The path to the underlying yaml file.
        self._config: ruamel.yaml.CommentedMap | None = None

    @property
    def config(self):
        if self._config is None:
            self._config = self.load()

        return AttrDict(self._config)

    @classmethod
    def load_from_path(cls, path: pt.Path) -> dict:
        """Read the configuration dictionary from disk."""
        try:
            with open(path, "r+") as cf:
                return yaml.load(cf)

        except FileNotFoundError as er:
            raise FileNotFoundError(
                f"Couldn't find the configuration file! Is it at {config_directory}? Error = {er.__repr__()}"
            )

    def load(self) -> dict:
        return self.__class__.load_from_path(self.path)

    def reload(self):
        """Reload the configuration file from disk."""
        self._config = None

    @classmethod
    def set_on_disk(cls, path: pt.Path | str, name: str | Collection[str], value: Any):
        _old = cls.load_from_path(path)

        if isinstance(name, str):
            name = name.split(".")
        else:
            pass

        setInDict(_old, name, value)

        with open(path, "w") as cf:
            yaml.dump(_old, cf)

    def set_param(self, name: str | Collection[str], value):
        self.__class__.set_on_disk(self.path, name, value)


pxconfig = YAMLConfiguration(config_directory)


def enforce_units(value: Any, preferred_units: u.Unit | str) -> u.Quantity:
    """
    Return a version of ``value`` with units of the type preferred or raise an error if that isn't possible.

    Parameters
    ----------
    value: any
        The value to enforce units on. Can be array or scalar with numerical type with / without units.
    preferred_units: :py:class:`astropy.units.Unit` or str
        The unit to enforce.

    Returns
    -------
    :py:class:`astropy.units.Quantity`
        The output quantity with the preferred units.

    """
    if isinstance(value, u.Quantity):
        try:
            return value.to(preferred_units)
        except u.UnitConversionError as error:
            raise ValueError(
                f"Failed to enforce units {preferred_units} on {value}. MSG: {error.__str__()}"
            )
    else:
        return value * u.Unit(preferred_units)


def getFromDict(dataDict: Mapping, mapList: Iterable[slice]) -> Any:
    """
    Fetch an object from a nested dictionary using a list of keys.

    Parameters
    ----------
    dataDict: dict
        The data dictionary to search.
    mapList: list
        The list of keys to follow.

    Returns
    -------
    Any
        The output value.

    """
    return reduce(operator.getitem, mapList, dataDict)


def setInDict(dataDict: Mapping, mapList: Iterable[slice], value: Any):
    """
    Set the value of an object from a nested dictionary using a list of keys.

    Parameters
    ----------
    dataDict: dict
        The data dictionary to search.
    mapList: list
        The list of keys to follow.
    value: Any
        The value to set the object to
    """
    getFromDict(dataDict, mapList[:-1])[mapList[-1]] = value


def rsetattr(obj: Any, attr: str, val: Any):
    """
    Recursively set an attribute.

    Parameters
    ----------
    obj: Any
        The object to search.
    attr: str
        The attribute position string.
    val: Any
        The value to set.
    """
    pre, _, post = attr.rpartition(".")
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)


def rgetattr(obj: Any, attr: str, *args) -> Any:
    """
    Recursively get an attribute.

    Parameters
    ----------
    obj: Any
        The object to search.
    attr: str
        The attribute position string.
    """

    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return functools.reduce(_getattr, [obj] + attr.split("."))


def produce_dict_heading(dataDict: Mapping, mapList: Iterable[slice]):
    """
    Generate iterative sub-dict in ``dataDict`` to allow setting of a sub-path for which the parent dict doesn't
    yet exist.

    Parameters
    ----------
    dataDict: dict
        The data dictionary to search.
    mapList: list
        The position to generate to.

    Returns
    -------
    None
    """
    for i, _ in enumerate(mapList, start=1):
        try:
            assert getFromDict(dataDict, mapList[:i]) is not None
        except Exception:
            setInDict(dataDict, mapList[:i], {})


def find_descriptors(cls: Any, descriptor_classes: tuple[Any]) -> dict[str, Any]:
    """Searches and returns all descriptors attached to this class or inherited."""
    _descriptor_instances = {
        k: v for k, v in cls.__dict__.items() if isinstance(v, descriptor_classes)
    }
    # iterate through all parent classes.
    _bases = cls.__bases__  # list of all parents.
    while object not in _bases:
        _nbases = []
        for _base in _bases:
            _descriptor_instances = {
                **_descriptor_instances,
                **{
                    k: v
                    for k, v in _base.__dict__.items()
                    if isinstance(v, descriptor_classes)
                },
            }
            _nbases += _base.__bases__
        _bases = _nbases[:]
    return _descriptor_instances


def merge_dicts(template: Mapping, partial: Mapping) -> Mapping:
    _output = {**template}  # Copy template.

    for k, v in template.items():
        if k in partial and v is not None:
            if isinstance(v, dict):
                _output[k] = merge_dicts(v, partial[k])
            else:
                _output[k] = partial[k]

        else:
            pass

    return _output



"""
Module containing core data structures and validation procedures for interacting with user data.

Notes
-----

``pyXMIP`` uses the ``pydantic`` library to validate user data. To do so, all of the data structures used in
the package require some boilerplate code to provide for validation. To do this without losing the readability of the
main package, this module contains all of that boilerplate code.

"""
from typing import Any, Callable, Mapping, Type

import astropy.coordinates as astro_coords
import numpy as np
from astropy.units import Quantity, Unit
from pydantic import (
    AfterValidator,
    BaseModel,
    ConfigDict,
    Field,
    GetCoreSchemaHandler,
    GetJsonSchemaHandler,
    model_validator,
)
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import PydanticUndefinedType, core_schema
from sqlalchemy.types import (
    BOOLEAN,
    DATETIME,
    FLOAT,
    INTEGER,
    NULLTYPE,
    TEXT,
    Interval,
    TypeEngine,
)
from typing_extensions import Annotated, Literal

from pyXMIP.utilities.logging import mainlog


def convert_np_type_to_sql(dtype: np.dtype) -> Type[TypeEngine]:
    """
    Convert a standard ``numpy`` data type into its corresponding ``sqlalchemy`` dtype.

    Parameters
    ----------
    dtype: :py:class:`numpy.dtype`
        The data type (numpy) to convert.

    Returns
    -------
    TypeEngine
        The class representing that data type in sqlalchemy.

    """

    _kind = dtype.kind

    if _kind == "f":
        return FLOAT
    elif _kind in ["i", "u"]:
        return INTEGER
    elif _kind in ["m"]:
        return Interval
    elif _kind in ["M"]:
        return DATETIME
    elif _kind in ["U", "S", "O"]:
        return TEXT
    elif _kind in ["b"]:
        return BOOLEAN
    else:
        return NULLTYPE


# ============================================================= #
# Custom Annotations
# ============================================================= #
class _UnitTypePydanticAnnotation:
    """
    This is a PyDantic annotation to process unit-typed objects.
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        """
        We return a pydantic_core.CoreSchema that behaves in the following ways:

        * strings will be parsed as `Unit` instances with the int as the x attribute
        * `Unit` instances will be parsed as `Unit` instances without any changes
        * Nothing else will pass validation
        * Serialization will always return just the string
        """

        def validate_from_str(value: str) -> Unit:
            result = Unit(value)
            return result

        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(validate_from_str),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=core_schema.union_schema(
                [
                    # check if it's an instance first before doing any further work
                    core_schema.is_instance_schema(Unit),
                    from_str_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda instance: str(instance)
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        # Use the same schema that would be used for `str`
        return handler(core_schema.str_schema())


class _QuantityTypePydanticAnnotation:
    """
    This is a PyDantic annotation to process quantity-typed objects.
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        """
        We return a pydantic_core.CoreSchema that behaves in the following ways:

        * strings will be parsed as `Unit` instances with the int as the x attribute
        * `Unit` instances will be parsed as `Unit` instances without any changes
        * Nothing else will pass validation
        * Serialization will always return just the string
        """

        def validate_from_str(value: str) -> Quantity:
            result = Quantity(value)
            return result

        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(validate_from_str),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=core_schema.union_schema(
                [
                    # check if it's an instance first before doing any further work
                    core_schema.is_instance_schema(Quantity),
                    from_str_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda instance: str(instance)
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        # Use the same schema that would be used for `str`
        return handler(core_schema.str_schema())


class _FrameTypePydanticAnnotation:
    """
    This is a PyDantic annotation to process coordinate frame objects.
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        def validate_from_str(value: str) -> astro_coords.GenericFrame:
            result = getattr(astro_coords, value)

            assert type(result) is type(
                astro_coords.GenericFrame
            ), f"Astropy coordinates {value} attribute is not a frame."
            return result

        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(validate_from_str),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=core_schema.union_schema(
                [
                    # check if it's an instance first before doing any further work
                    core_schema.is_instance_schema(astro_coords.GenericFrame),
                    from_str_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda instance: instance.__name__
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        # Use the same schema that would be used for `str`
        return handler(core_schema.str_schema())


class _CMDTypePydanticAnnotation:
    """
    This is a PyDantic annotation to process :py:class:`CrossMatchDatabase` objects.
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        from pyXMIP.cross_reference import CrossMatchDatabase

        def validate_from_str(value: str) -> CrossMatchDatabase:
            result = CrossMatchDatabase(value)
            return result

        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(validate_from_str),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=core_schema.union_schema(
                [
                    # check if it's an instance first before doing any further work
                    core_schema.is_instance_schema(CrossMatchDatabase),
                    from_str_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda instance: instance.__name__
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        # Use the same schema that would be used for `str`
        return handler(core_schema.str_schema())


# ============================================================= #
# Custom Validators
# ============================================================= #
PydanticUnit = Annotated[Unit, _UnitTypePydanticAnnotation]
# Any: The Pydantic type representing the :py:class:`astropy.units.core.Unit` class.

PydanticQuantity = Annotated[Quantity, _QuantityTypePydanticAnnotation]
# Any: The Pydantic type representing the :py:class:`astropy.units.core.Quantity` class.

PydanticFrame = Annotated[astro_coords.GenericFrame, _FrameTypePydanticAnnotation]
# Any: The Pydantic type representing the :py:class:`astropy.units.core.Quantity` class.


def construct_template(datamodel: Type[BaseModel]) -> dict:
    """
    Construct a template version of a known data model. This is largely useless for external users,
    but is used in the backend to produce model schema which can then be filled in with information from objects like tables.

    Parameters
    ----------
    datamodel: BaseModel
        The ``pydantic`` model to produce a template version of.

    Returns
    -------
    dict
        The output template.
    """
    structure = {
        k: None for k, v in datamodel.model_fields.items()
    }  # Grab all of the model fields.

    for k, v in datamodel.model_fields.items():
        if type(v.default) is not PydanticUndefinedType and v.default is not None:
            structure[k] = v.default
        else:
            _underlying_type = v.annotation

            if type(_underlying_type) is type(BaseModel):
                structure[k] = construct_template(_underlying_type)
            else:
                pass

    return structure


# ============================================================= #
# Custom Structures
# ============================================================= #
#   Custom structures are Pydantic recognized data structures for validation
#   and eventual usage in schema objects.


class Registry(dict):
    """
    Bare-bones registry class.
    """

    def __init__(self, mapping: Mapping):
        """
        Initialize the generalized registry class.

        Parameters
        ----------
        mapping: dict
            The dictionary underlying this registry.
        """
        super().__init__(mapping)

    def as_list(self) -> list:
        return list(self.values())

    def add(self, key: str, value: Any):
        assert key not in self, f"Key {key} already in registry."
        self[key] = value


class TableColumn(BaseModel):
    """
    Pydantic data class representing a single column in a table.

    The :py:class:`TableColumn` class is utilized by :py:class:`ColumnMap` to construct the column structure for
    :py:class:`schema.SourceTableSchema`.
    """

    # -- Managing configuration -- #
    model_config = ConfigDict(validate_assignment=True)

    name: str
    """str: The column name corresponding to this column."""
    unit: PydanticUnit | None = None
    """The unit to attribute to this column."""


class SourceTableSchemaSettings(BaseModel):
    """
    Pydantic data class representing the settings for a :py:class:`schema.SourceTableSchema`.
    """

    # -- Managing configuration -- #
    model_config = ConfigDict(validate_assignment=True)

    # ============================= #
    # Settings                      #
    # ============================= #
    default_coord_system: Literal["ICRS", "Galactic"] = "ICRS"
    """str: The default coordinate system to use.

    Use this to select the native coordinate system for the schema. Whichever coordinate system is selected, there must
    be sufficient identifiable columns to actually obtain those coordinates.

    """
    object_type_separator: str = Field(default="|", min_length=1, max_length=1)
    """str: The separator character / string between different object types.

        In many databases, returned object types may have multiple types separated by a delimiter. This delimiter should
        be specified here if it is needed.

        .. important::

            In all native ``pyXMIP`` data objects, a "|" is used for the delimiter. Even if the object type is a single
            value, all object types in ``pyXMIP`` tables are formatted ``"|entry_1|entry_2|...|"``.
    """


class ColumnMap(BaseModel):
    """
    Pydantic data class representing a mapping from columns in a table to standard pyxmip columns.

    The :py:class:`ColumnMap` class is used to construct the column structure for :py:class:`schema.SourceTableSchema`.
    """

    # ========================================================================== #
    # Validators                                                                 #
    # ========================================================================== #
    # -- Managing configuration -- #
    model_config = ConfigDict(validate_assignment=True)

    @staticmethod
    def validate_table_column_unit(
        expected_unit: Unit, equivalencies: list[tuple] | None = None
    ):
        """
        PyDactic validator method to enforce unit norms on :py:class:`TableColumn` instances in the :py:class:`ColumnMap`.

        Parameters
        ----------
        expected_unit: Unit
            The expected unit. The unit specified in the data must be convertible to this unit.
        equivalencies
            The equivalencies to allow when checking for type consistency. By default, ``None``.

        Raises
        ------
        ValidationError
            [UNIT CONSISTENCY FAULT] Failed to convert ``value`` to units ``expected_unit``.

        """

        def _check_unit_validation(value: TableColumn) -> TableColumn:
            # -- Check if NONE -- #
            # If the unit isn't provided, we can just assign the expected unit.
            if value.unit is None:
                value.unit = expected_unit

            # -- Check unit type -- #
            # The user might have specified a string, convert to unit.
            if isinstance(value.unit, str):
                value.unit = Unit(value.unit)

            try:
                _ = value.unit.to(expected_unit, equivalencies=equivalencies)
            except Exception:
                raise AssertionError(
                    f"[UNIT CONSISTENCY FAULT] Failed to convert {value} to units {expected_unit}. Raised by ColumnMap.validate_table_column_unit."
                )
            return value

        return _check_unit_validation

    @staticmethod
    def validate_table_column_no_unit(value: TableColumn) -> TableColumn:
        """
        PyDactic validator method to enforce unit norms on :py:class:`TableColumn` instances in the :py:class:`ColumnMap`.

        Parameters
        ----------
        value
            The :py:class:`TableColumn` being validated.


        Raises
        ------
        ValidationError
            [UNIT CONSISTENCY FAULT] Failed to convert ``value`` to units ``expected_unit``.

        """
        assert (
            value.unit is None
        ), f"[UNIT CONSISTENCY FAULT] This TableColumn should not have units but is {value}. Raised by ColumnMap.validate_table_column_no_unit."
        return value

    # Coordinate Columns #
    RA: Annotated[
        TableColumn, AfterValidator(validate_table_column_unit(Unit("deg")))
    ] | None = None
    """The column containing the object RA coordinate."""
    DEC: Annotated[
        TableColumn, AfterValidator(validate_table_column_unit(Unit("deg")))
    ] | None = None
    """The column containing the object DEC coordinate."""
    RA_ERR: Annotated[
        TableColumn, AfterValidator(validate_table_column_unit(Unit("deg")))
    ] | None = None
    """The column containing the object RA coordinate error."""
    DEC_ERR: Annotated[
        TableColumn, AfterValidator(validate_table_column_unit(Unit("deg")))
    ] | None = None
    """The column containing the object DEC coordinate error."""
    GAL_L: Annotated[
        TableColumn, AfterValidator(validate_table_column_unit(Unit("deg")))
    ] | None = None
    """The column containing the object's galactic longitude."""
    GAL_B: Annotated[
        TableColumn, AfterValidator(validate_table_column_unit(Unit("deg")))
    ] | None = None
    """The column containing the object's galactic latitude."""
    GAL_L_ERR: Annotated[
        TableColumn, AfterValidator(validate_table_column_unit(Unit("deg")))
    ] | None = None
    """The column containing the object's galactic longitude error."""
    GAL_B_ERR: Annotated[
        TableColumn, AfterValidator(validate_table_column_unit(Unit("deg")))
    ] | None = None
    """The column containing the object's galactic latitude error."""
    SEP: Annotated[
        TableColumn, AfterValidator(validate_table_column_unit(Unit("deg")))
    ] | None = None
    # Name columns #
    NAME: Annotated[TableColumn, AfterValidator(validate_table_column_no_unit)]
    """The column containing the object's main identifier."""
    # Extra Columns #
    Z: Annotated[
        TableColumn, AfterValidator(validate_table_column_no_unit)
    ] | None = None
    """The column containing the object's redshift."""
    TYPE: Annotated[
        TableColumn, AfterValidator(validate_table_column_no_unit)
    ] | None = None
    """The column containing the object type of the catalog object."""

    def search_inverse(self, key):
        return {v: k for k, v in self.model_fields.items()}[key]


class ColumnStdErrorSpecifier(BaseModel):
    """
    Pydantic data class specifying the error on a given column of a table.

    .. note::

        This class is table-agnostic, it won't know anything about the data it's describing until runtime. Thus,
        failure to create the class accurately may lead to late appearing validation errors because i.e. your table doesn't
        have the specified column.

    Notes
    -----

    The :py:class:`ColumnErrorSpecifier` should be understood as an "archetypal" structure for any reduction process where
    the precision in a particular field of the matching database is used. Users must specify a ``base_column``, corresponding to
    the column (i.e. RA) for which this specifier is determining the error (i.e. RA_PREC). Error can either by dynamic, as determined by
    providing an ``error_column``, or by providing an ``error_value``.
    """

    base_column: str | None = None
    """:py:class:`TableColumn`: The name of the column who's error is being specified.

    This should be the name of a column in the table which is being described by this data object.
    """
    error_column: TableColumn | None = None
    """:py:class:`TableColumn` optional: The column which specifies the error in :py:attr:`ColumnErrorSpecifier.base_column`

    At least one of :py:attr:`ColumnErrorSpecifier.error_column` and :py:attr:`ColumnErrorSpecifier.error_value` is required
    for validation to pass without errors.
    """
    error_value: PydanticQuantity | None = None
    """:py:class:`astropy.units.core.Quantity`, optional: The global value of the error in :py:attr:`ColumnErrorSpecifier.base_column`

    .. hint::

        If a table doesn't have different errors for different entries but instead has a flat error profile, this is the
        option you should use.

    At least one of :py:attr:`ColumnErrorSpecifier.error_column` and :py:attr:`ColumnErrorSpecifier.error_value` is required
    for validation to pass without errors.
    """
    error_convert_to_std: Callable = lambda x: x
    """Callable, optional: Optional parameter to provide a function to convert error to standard deviation.

    This class **assumes** that the provided error is the standard deviation of a normally distributed error. If the available
    error column is not the standard deviation, specify this function to take ``x`` (the uncorrected error) as an argument and
    return the standard deviation form of the error.
    """

    def validate_on_table(self, table):
        """
        Validate this column error specifier against a table.

        Parameters
        ----------
        table: :py:class:`structures.table.SourceTable`
            The table to check this data structure against.

        Notes
        -----

        This validator performs 3 checks:

        1. Assert that ``self.base_column`` is actually a column in the table.
        2. If ``self.error_column`` is specified, then it must also be in the table.
        3. At least one of ``self.error_column`` and ``self.error_value`` must be specified.

        """

        if self.error_column is not None:
            assert (
                self.error_column.name in table.columns
            ), f"The base column {self.error_column.name} is not present in the validating table."

        assert not (
            self.error_column is None and self.error_value is None
        ), "At least one of error_column or error_value must be specified."

    @model_validator(mode="after")
    def validator(self):
        """
        Validate this column error specifier.


        Notes
        -----

        This validator performs 1 check:

        1. At least one of ``self.error_column`` and ``self.error_value`` must be specified.

        """
        # Check that at least one of the needed columns is specified.
        assert not (
            self.error_column is None and self.error_value is None
        ), "At least one of error_column or error_value must be specified."

        return self


class ICRSCoordinateStdErrorSpecifier(BaseModel):
    """
    Pydantic data class specifying the astrometric errors in a table.

    Notes
    -----

    .. hint::

        Like many of the structures in :py:mod:`utilities.types`, this class is composed of various subtypes. When documentation
        says that something must have the "format" of another class, that implies that it may **either** be an instance of the
        actual class, or the ``dict`` provided must be interpretable as that class.

    To initialize a :py:class:`CoordinateErrorSpecifier`, you need to provide the following:

    - :py:attr:`CoordinateErrorSpecifier.longitude_column`: The column in the table representing the longitude (i.e. RA).
    - :py:attr:`CoordinateErrorSpecifier.latitude_column`: The column in the table representing the latitude (i.e. DEC).

    Both of these should have the native format of :py:class:`TableColumn`.

    .. warning::

        The two coordinate columns must be from the same coordinate system as ``pyXMIP`` treats them as generic lat/lon instead
        of trying to interpret



    """

    # Error specifiers #
    lon_error: TableColumn | PydanticQuantity | None = None
    """:py:class:`TableColumn` or :py:class:`astropy.units.core.Quantity`: The error specification in the longitude."""
    lat_error: TableColumn | PydanticQuantity | None = None
    """:py:class:`TableColumn` or :py:class:`astropy.units.core.Quantity`: The error specification in the latitude."""
    position_error: TableColumn | PydanticQuantity | None = None
    """:py:class:`TableColumn` or :py:class:`astropy.units.core.Quantity`: The circular error specification."""

    @property
    def mode(self):
        if self.position_error is not None:
            return "circular"
        elif all([cerr is not None for cerr in [self.lat_error, self.lon_error]]):
            return "axial"
        else:
            return None

    @model_validator(mode="after")
    def validator(self):
        """
        Validate the :py:class:`CoordinateErrorSpecifier`.

        Notes
        -----

        The only check performed by this validator is that either the lat/lon error is specified or the circular is specified.
        """
        if all(
            [i is None for i in [self.lat_error, self.lon_error, self.position_error]]
        ):
            return self

        if self.position_error is None:
            assert not any(
                cerr is None for cerr in [self.lat_error, self.lon_error]
            ), f"CoordinateErrorSpecifier {self} is not circular but doesn't have 2 axes for position error."

        else:
            for cerr in [self.lat_error, self.lon_error]:
                if cerr is not None:
                    mainlog.warning(
                        f"CoordinateErrorSpecifier {self} is in MODE=circular, but axial errors are specified: {cerr}."
                    )

        return self

    def check_empty(self):
        assert any(
            [
                i is not None
                for i in [self.lat_error, self.lon_error, self.position_error]
            ]
        ), "At least one error specifier is needed."
